/*
COPYMID -- Copy part of a text file by string, byte offset or line offset
    Copyright (C) 2003  Norm London, London Computing

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	norm_london@londoncomputing.com

    Compile with Visual Studio 6 or later.  Requires comdlg32.lib odbc32.lib odbccp32.lib winmm.lib
	cl xfer32.c /link comdlg32.lib odbc32.lib odbccp32.lib winmm.lib

*/

#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#ifdef WIN32
#include <fcntl.h>
#include <io.h>
#include <stdlib.h>
#include <process.h>
#include <windows.h>
#endif

#define LINESI 4096
#define BUFLEN 65000
char line[LINESI],trline[LINESI],startp[LINESI],endp[LINESI],buf[BUFLEN],outbuf[BUFLEN];

void eputs(s)
char *s;
{
	if (fputs(s,stdout)) {
		perror("Write Error");
		exit(4);
	}
}

#define LMAX	512
#define PMAX	512
#define ANY	    '?'
#define CLASS	'['
#define ENDCLASS ']'
#define ESCAPE  '\\'
#define STAR	'*'
#define RANGE '-'
#define ENDPAT  0
char	lbuf[LMAX];
char	pbuf[PMAX];
/*******************************************************/
char	*pp;
#ifdef TEST
#include <stdio.h>
main(argc,argv)
int argc;
char *argv[];
{
	printf("%s %s pattern %s\n",argv[1],uxmatch(argv[1],argv[2]) ? "matches" : "does not match",argv[2]);
}
#endif

int uxmatch(line, pattern)
char	*line;     /* (partial) line to match      */
char	*pattern;  /* (partial) pattern to match   */
{
	register char   *l;	    /* Current line pointer	      */
	register char   *p;	    /* Current pattern pointer      */
//	char *endclass;		    /* End of class pointer	*/
//	register char   c;	    /* Current character	      */
	int		   op;	      	/* Pattern operation	      */
//	int		   n;	      	/* Class counter		      */
	char inclass = 0, lowvalue = 0, foundmatch = 0, escape = 0;
	l = line;
	p = pattern;
	if (*p == ENDPAT)
		return 1;

	while ((op = *p++) != ENDPAT) {
		escape = 0;
		if (op == '\\') {
			op = *p++;
			if (op)		
				escape = 1;
			else {	/* pattern ends with \. look for \ */
				--p;
				op = '\\';
			}
		}

		if (inclass) {  
			/* currently scanning a class spec [abc-de] */
			if (!foundmatch) {
				/* and we haven't found a match yet */
				if (lowvalue) {
					/* but we are in a range c-e */
					/* op is high value of range */
					if (tolower(*l) >= lowvalue && tolower(*l) <= tolower(op)) {
						/* we found a match */
						foundmatch = 1;
						lowvalue = 0;
						continue;
					}
					/* no match found, return failure */
					return 0;
				}

				/* not in a range */
				if (*p == RANGE)  {
					/* in a range, op is low value */
					lowvalue = tolower(op);
					++p;
					continue;
				}
				else
					lowvalue = 0;


				if (tolower(*l) == tolower(op)) {
					/* found match, whoopie */
					foundmatch = 1;
					continue;
				}
			}
			if (op == ENDCLASS && !escape) {
				if (!foundmatch)					
					return 0;		/* not found, sorry charlie */
				inclass = 0;
				++l;
			}
		}
		else {
			switch (op) {
			case ANY:
				if (*l++ == '\0')
					return(0);
				break;
			case CLASS:
				inclass = 1;
				lowvalue = 0;
				foundmatch = 0;
				break;				
			case STAR:		 /* Zero or more characters	*/
				/* match line against rest of pattern */
				while (*l)
					if (uxmatch(l++,p))
						return 1;
				return 0;		 /* Nothing else worked */
			default:
				if (tolower(*l) != tolower(op))
					return(0);
				l++;
				break;
			}
		}
	}
	return *l == 0 ? 1 : 0;   /* if everything matches and at end of string, ok */
}


void feputs();

struct _stati64 statbuf;

int copybytes(char *s_offset, char *s_count) {
	_int64 offset = -1, count = 0x7fffffffffffffff, tcount;
	unsigned kt=1;

#ifdef WIN32
	setmode(fileno(stdin),O_BINARY);
	setmode(fileno(stdout),O_BINARY);
#endif

	offset = _atoi64(s_offset);
	if (s_count)
		tcount = _atoi64(s_count);
	else
		tcount = 0;

	if (tcount > 0)
		count = tcount;

	if (offset)
		/* start at specified offset into file, from 0 */
		_lseeki64(_fileno(stdin),offset,0);

	if (feof(stdin)) {
		fprintf(stderr,"End of file reached before starting byte.\n");
		exit(3);
	}

	while(count > 0 && kt > 0) {
		/* copy to count bytes or end of file */
		kt = fread(buf,1,BUFLEN,stdin);
		if (count < kt)
			kt = (unsigned)count;
		if (kt)
			fwrite(buf,1,kt,stdout);
		count -= kt;
	}

	if (ferror(stdin)) {
		fprintf(stderr,"Read Error.\n");
		exit(4);
	}

	if (count && count != 0x7fffffffffffffff) {
		fprintf(stderr,"End of file reached before ending byte.\n");
		exit(1);
	}
		
	exit(0);
}

int copylines(char *s_offset, char *s_count) {
	long offset = -1,i;
	size_t kt,count = 0x7fffffff, tcount;

	offset = atoi(s_offset);

	if (s_count)
		tcount = atoi(s_count);
	else
		tcount = 0;

	if (tcount > 0)
		count = tcount;

	for (i=1;i < offset && !feof(stdin) && !ferror(stdin);++i) 
		/* start at specified line number */
		fgets(buf,sizeof(buf),stdin);

	if (i < offset) {
		if (feof(stdin))
			fprintf(stderr,"Reached end of file before starting line.\n");
		else
			fprintf(stderr,"Read Error while positioning to starting line.\n");
		exit(3);
	}

	for (kt=0;kt<count && !feof(stdin) && !ferror(stdin);++kt) {
		/* copy to count bytes or end of file */
		if (fgets(buf,sizeof(buf),stdin))
			fputs(buf,stdout);
	}

	if (ferror(stdin)) {
		fprintf(stderr,"Read Error while copying.\n");
		exit(4);
	}

	if (kt < count && count != 0x7fffffff) {
		fprintf(stderr,"End of File reached before ending line.\n");
		exit(1);
	}

	exit(0);
}

void main(int argc,char *argv[]) {

	char *li,iflag,copyeof,uflag, *count=NULL;
	int i;
	FILE *hlp;

#ifdef WIN32
	char tmp[_MAX_PATH],*p,cmd[_MAX_PATH];
	if ((p=getenv("tmp")) == NULL)
		if ((p = getenv("temp")) == NULL)
			GetWindowsDirectory(tmp,_MAX_PATH);
		else
			strcpy(tmp,p);
	else
		strcpy(tmp,p);
	if (tmp[strlen(tmp)-1] != '\\')
		strcat(tmp,"\\");

	strcat(tmp,"copymid.hlp");
	hlp = fopen(tmp,"w");
#else
	hlp = stderr;
#endif


	/* argv[1] is source; otfi is dest;
	 argv[3] is start str; argv[4] is end string */
	 if (argc < 3) {

		fprintf(hlp,"\nCopyright 2003, Norm London, London Computing.\nnorm_london@londoncomputing.com\n\n");
        fprintf(hlp,"This program is free software; you can redistribute it and/or modify\n");
		fprintf(hlp,"it under the terms of the GNU General Public License as published by\n");
		fprintf(hlp,"the Free Software Foundation; either version 2 of the License, or\n");
		fprintf(hlp,"(at your option) any later version.\n\n");
		fprintf(hlp,"This program is distributed in the hope that it will be useful,\n");
		fprintf(hlp,"but WITHOUT ANY WARRANTY; without even the implied warranty of\n");
		fprintf(hlp,"MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n");
		fprintf(hlp,"GNU General Public License for more details.\n\n");
		fprintf(hlp,"You should have received a copy of the GNU General Public License\n");
		fprintf(hlp,"along with this program; if not, write to the Free Software\n");
		fprintf(hlp,"Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA\n\n");


	 	fprintf(hlp,"Usage is:\n");
	 	fprintf(hlp,"COPYMID [-I] [-U] starting-pattern ending-pattern [<in] [>out] -or- \n");
	 	fprintf(hlp,"COPYMID -L starting-line [linecount] [<in] [>out] -or- \n");
	 	fprintf(hlp,"COPYMID -B starting-byte [bytecount] [<in] [>out]  \n");
	 	fprintf(hlp,"Copies lines from standard input to standard output\n");
		fprintf(hlp,"If -L specified, starts copying at starting-line and copies linecount\n");
		fprintf(hlp,"lines.  If linecount is not present, copy to end of file.\n\n");

		fprintf(hlp,"If -B specified, starts copying at Byte# starting-byte for bytecount\n");
		fprintf(hlp,"bytes.  If bytecount is not present, copy to end of file.\n\n");

		fprintf(hlp,"If neither -L or -B specified, starts copying with the line following the.\n");
	 	fprintf(hlp,"first line equal to the starting-pattern.  Copies up to, but not including the\n");
	 	fprintf(hlp,"line ending-pattern. If starting pattern is '-start', starts with the\n");
	 	fprintf(hlp,"first line.  If ending pattern is '-end' or not found, copies to end\n");
	 	fprintf(hlp,"of file.  Patterns must match exactly, but may contain wildcard \n");
	 	fprintf(hlp,"characters '*' and '?'.  '*' stands for any 0 or more characters.\n");
	 	fprintf(hlp,"'?' stands for exactly one character.  Patterns should be enclosed in\n");
	 	fprintf(hlp,"quotes (\") if they contain any punctuation characters or spaces.  They\n");
	 	fprintf(hlp,"should use the following special symbols, when needed:\n");
	 	fprintf(hlp,"\\\\ -- BackSlash; \\\" -- Double Quote\n\n");
	 	fprintf(hlp,"Maximum line size is %d characters.\n",LINESI);
	 	fprintf(hlp,"If -I specified, starting and ending lines are included.\n");
	 	fprintf(hlp,"If -U specified, case is ignored during pattern matching.\n");
	 	fprintf(hlp,"Returns 0--Good Copy,     1--Ending String Not Found\n");
	 	fprintf(hlp,"        2--Usage Display, 3--Starting String Not Found\n");
	 	fprintf(hlp,"        4--Write Error\n");


#ifdef WIN32
		fclose(hlp);
		strcpy(cmd,"NOTEPAD.EXE ");
		strcat(cmd,tmp);
		WinExec(cmd,SW_SHOW);
#endif
	 	exit(2);
	 }

	_fstati64(fileno(stdin),&statbuf);
	if (statbuf.st_size == 0) {
		fprintf(stderr,"Input not redirected or file is empty.\n");
		exit(0);
	}


	uflag = iflag = 0;

	for (i = 1; i < argc-1; ++i) {
		if (argv[i][0] == '-') {
			switch(toupper(argv[i][1])) {
			case 'I': 
				iflag = 1; 
				break;
			case 'U': 
				uflag = 1; 
				break;
			case 'B': // next two args are offset and optional count
				if (argc > (i+2))
					count = argv[i+2];
				copybytes(argv[i+1],count);
				break;
			case 'L': // next two args are starting line and line count
				if (argc > (i+2))
					count = argv[i+2];
				copylines(argv[i+1],count);
				break;
			}
		}
	}
	 
	 strcpy(startp,argv[1+uflag+iflag]);
	 strcpy(endp , argv[2+uflag+iflag]);
	 
 	setvbuf(stdin,buf,_IOFBF,BUFLEN);

	setvbuf(stdout,outbuf,_IOFBF,BUFLEN);

	strcpy(line,endp);
	strupr(line);
	copyeof = (char)(!(strcmp(line,"-END")));

	strcpy(line,startp);
	strupr(line);
	li = NULL;

	if (uflag) {
		strupr(startp);
		strupr(endp);
	}

	if (strcmp(line,"-START")) {
  /* if looking for a starting line, get to it */
		while((li = gets(line)) != NULL) {
			if (uflag) {
				strcpy(trline,line);
				strupr(trline);
	 			if (uxmatch(trline,startp))
	 				break;
	 		}
	 		else
	 			if (uxmatch(line,startp))
	 				break;
	 	}
		if (li == NULL) {
		 	fprintf(stderr,"%s <=== Starting String Not Found!\n",startp);
	 		exit(3);
	 	}
	 	else if (iflag)
	 		eputs(line);
	}
		 
	 while((li = fgets(line,sizeof(line),stdin)) != NULL) {

	 	if (!copyeof) 
	 		if (uflag) {
		 		strcpy(trline,line);
		 		strupr(trline);
			 	if (uxmatch(trline,endp))
			 		break;
			 }
			 else
			 	if (uxmatch(line,endp))
			 		break;

	 	eputs(line);
	 }
	 
	if (li == NULL) {
	 	if (!copyeof) {
		 	fprintf(stderr,"Ending string: %s, not found.\n",endp);
		 	exit(1);
		 }
	}
	else if (iflag)
	 	eputs(line);
	fflush(stdout);
	exit(0);
}
	 

